from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from django_alt.abstract.endpoints import MetaEndpoint
from django_alt.endpoints import Endpoint
from django_alt_tests.conf.endpoints import ModelASerializer, ModelAEndpoint1
from django_alt_tests.conf.models import ModelA


class MetaEndpointTests(TestCase):
    def test_transform_config(self):
        def are_equal(c1, c0):
            c1 = MetaEndpoint.transform_config_shorthands(c1)
            self.assertDictEqual(c0, c1)

        are_equal({
            'get': None
        }, {
            'get': {}
        })

        are_equal({
            'get, post': None
        }, {
            'get': {},
            'post': {}
        })

        are_equal({
            ' get,post, put': None
        }, {
            'get': {},
            'post': {},
            'put': {},
        })
        are_equal({
            ' get,post': {
                'q': 1
            }
        }, {
            'get': {'q': 1},
            'post': {'q': 1},
        })
        are_equal({
            ' get,post': {
                'q': 1
            },
            'post': {
                'z': 2
            }
        }, {
            'get': {'q': 1},
            'post': {'q': 1, 'z': 2},
        })
        are_equal({
            ' get': {
                'q': 1
            },
            'get': {
                'z': 2
            }
        }, {
            'get': {'q': 1, 'z': 2},
        })
        are_equal({
            ' get, get': {
                'q': 1
            },
            'post,get': {
                'z': 4
            }
        }, {
            'get': {'q': 1, 'z': 4},
            'post': {'z': 4}
        })

    def test_unknown_http_method(self):
        with self.assertRaises(AssertionError):
            class MyEndpoint(Endpoint):
                serializer = ModelASerializer
                config = {'get,foo': None}


class EndpointTests(TestCase):
    def setUp(self):
        self.client = APIClient()

    def test_required_fields(self):
        with self.assertRaises(AssertionError):
            class MyEndpoint(Endpoint):
                pass

        with self.assertRaises(AssertionError):
            class MyEndpoint1(Endpoint):
                config = {
                    'foo': {
                        'query': None
                    }
                }

        with self.assertRaises(AssertionError):
            class MyEndpoint2(Endpoint):
                config = {
                    'get': {
                        'query': None
                    }
                }

        with self.assertRaises(AssertionError):
            class MyEndpoint3(Endpoint):
                config = {
                    'get': {
                        'filters': {
                            'y': lambda x, y: (x, y)
                        }
                    }}

        with self.assertRaises(AssertionError):
            class MyEndpoint4(Endpoint):
                config = {
                    'get': {
                        'query': lambda x, y: (x, y),
                        'filters': {
                            'x': lambda x, y: (x, y)
                        }
                    }}

        with self.assertRaises(AssertionError):
            class MyEndpoint5(Endpoint):
                serializer = ModelASerializer
                config = {
                    'get': {
                        'query': lambda x, y: (x, y),
                        'filters': lambda x, y: (x, y)
                    }}

        with self.assertRaises(AssertionError):
            class MyEndpoint7(Endpoint):
                serializer = ModelASerializer
                config = {
                    'get': {
                        'query': {'something_strange'},
                        'filters': {
                            'x': lambda x, y: (x, y)
                        }
                    }}

        with self.assertRaises(AssertionError):
            class MyEndpoint8(Endpoint):
                serializer = ModelASerializer
                config = {
                    'delete': None
                }

        class MyEndpoint6(Endpoint):
            serializer = ModelASerializer
            config = {
                'get': {
                    'query': lambda x, y: (x, y),
                    'filters': {
                        'x': lambda x, y: (x, y)
                    }
                },
                'delete': {
                    'query': lambda x, y: (x, y)
                }}

    def test_basic_get(self):
        self.assertTrue(hasattr(ModelAEndpoint1, 'as_view'))
        resp = self.client.get(reverse('e1'))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.data, [])

        ModelA.objects.create(field_1='aa', field_2=5)
        resp = self.client.get(reverse('e1'))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(dict(resp.data[0]), {'id': 1, 'field_1': 'aa', 'field_2': 5})

    def test_method_not_supported(self):
        self.assertEqual(self.client.post(reverse('e1')).status_code, 405)
        self.assertEqual(self.client.patch(reverse('e1')).status_code, 405)
        self.assertEqual(self.client.put(reverse('e1')).status_code, 405)
        self.assertEqual(self.client.delete(reverse('e1')).status_code, 405)

    def test_raises_validation_error(self):
        ModelA.objects.create(field_1='aa', field_2=5)
        resp = self.client.get(reverse('e2'))
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(dict(resp.data[0]), {'id': 1, 'field_1': 'aa', 'field_2': 5, 'extra': 'this is an extra'})

        ModelA.objects.create(field_1='aa', field_2=5)
        resp = self.client.get(reverse('e2'))
        self.assertEqual(resp.status_code, 400)
        self.assertDictEqual(dict(resp.data), {'error': ['msg.']})

    def test_filtered_get(self):
        ModelA.objects.create(field_1='aaa', field_2=5)
        ModelA.objects.create(field_1='aaa', field_2=6)
        ModelA.objects.create(field_1='aca', field_2=10)
        ModelA.objects.create(field_1='ada', field_2=10)
        ModelA.objects.create(field_1='aea', field_2=10)
        ModelA.objects.create(field_1='aaa', field_2=10)

        resp = self.client.get(reverse('e3') + '?filter1=aaa')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.data), 3)
        for d in resp.data:
            self.assertEqual(d['field_1'], 'aaa')

        resp = self.client.get(reverse('e3') + '?filter2=10')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.data), 4)
        for d in resp.data:
            self.assertEqual(d['field_2'], 10)

        resp = self.client.get(reverse('e3') + '?filter1=aaa&filter2=10')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.data), 1)
        for d in resp.data:
            self.assertEqual(d['field_1'], 'aaa')
            self.assertEqual(d['field_2'], 10)

    def test_get_precondition_permission_check(self):
        resp = self.client.get(reverse('e4'))
        self.assertEqual(resp.status_code, 401)

    def test_get_postcondition_permission_check(self):
        resp = self.client.get(reverse('e5'))
        self.assertEqual(resp.status_code, 401)

    def test_basic_post(self):
        resp = self.client.post(reverse('e6'), {'field_1': 'aaa', 'field_2': 3})
        self.assertEqual(resp.status_code, 201)
        self.assertDictEqual(resp.data, {'id': 1, 'field_1': 'aaa', 'field_2': 3})
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.all().first().field_1, 'aaa')
        self.assertEqual(ModelA.objects.all().first().field_2, 3)

    def test_basic_patch(self):
        ModelA.objects.create(field_1='aaa', field_2=3)
        resp = self.client.patch(reverse('e6'), {'field_1': 'bbb'})
        self.assertEqual(resp.status_code, 200)
        self.assertDictEqual(resp.data, {'id': 1, 'field_1': 'bbb', 'field_2': 3})
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.all().first().field_1, 'bbb')
        self.assertEqual(ModelA.objects.all().first().field_2, 3)

    def test_basic_delete(self):
        ModelA.objects.create(field_1='aca', field_2=10)
        ModelA.objects.create(field_1='ada', field_2=10)
        ModelA.objects.create(field_1='aea', field_2=10)

        resp = self.client.delete(reverse('e7'))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(ModelA.objects.count(), 2)

    def test_delete_many(self):
        ModelA.objects.create(field_1='aca', field_2=10)
        ModelA.objects.create(field_1='ada', field_2=10)
        ModelA.objects.create(field_1='aea', field_2=10)

        resp = self.client.delete(reverse('e8'))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(ModelA.objects.count(), 0)

    def test_fields_from_url(self):
        resp = self.client.post(reverse('e9', kwargs={'field_1': 'abc', 'field_2': 123}))

        self.assertEqual(resp.status_code, 201)
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().field_1, 'abc')
        self.assertEqual(ModelA.objects.first().field_2, 123)

    def test_fields_from_url_override_args(self):
        resp = self.client.post(reverse('e9', kwargs={'field_1': 'abc', 'field_2': 123}), {
            'field_1': 'def'
        })

        self.assertEqual(resp.status_code, 201)
        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().field_1, 'abc')
        self.assertEqual(ModelA.objects.first().field_2, 123)

    def test_fields_from_url_nonexistent_args(self):
        with self.assertRaises(AssertionError) as e:
            self.client.post(reverse('e10', kwargs={'field_1': 'abc', 'field_2': 123}), )



def compose_and(*functions) -> bool:
    """
    Composes an iterable of callables with identical signatures
    and asserts universal quantification for their results.
    :param functions: ant iterable of functions to call
    :return: whether all function returns evaluate to true
    """
    def func_prototype(*args, **kwargs):
        return all(f(*args, **kwargs) for f in functions)
    return func_prototype



from unittest.mock import patch

from django.test import TestCase
from rest_framework import serializers

from django_alt.abstract.serializers import BaseValidatedSerializer
from django_alt.abstract.validators import Validator
from django_alt.serializers import ValidatedModelSerializer
from django_alt.utils.shortcuts import invalid, invalid_if, if_in
from django_alt_tests.conf.models import ModelA


def generate_serializer(validator_class):
    cls = validator_class

    class ConcreteSerializer(BaseValidatedSerializer):
        somestring = serializers.CharField()

        class Meta:
            validator_class = cls

        def create(self, validated_data):
            self.Meta.validator_instance.did_create(validated_data, validated_data)
            return validated_data

    return ConcreteSerializer


def serialize(validator, data):
    serializer = generate_serializer(validator)(data=data)
    serializer.is_valid(raise_exception=True)
    return serializer.save()


class BaseValidatedSerializerTests(TestCase):
    def setUp(self):
        self.validator_class = type('ConcreteValidator', (Validator,), dict())

    def test_init_with_no_validator_class(self):
        class ConcreteSerializer(BaseValidatedSerializer): pass

        with self.assertRaises(AssertionError):
            inst = ConcreteSerializer()

    def test_init_with_validator_class(self):
        class ConcreteSerializer(BaseValidatedSerializer):
            pass

        inst = ConcreteSerializer(validator_class=self.validator_class)
        self.assertTrue(hasattr(inst, 'Meta'))
        self.assertTrue(hasattr(inst.Meta, 'validator_class'))
        self.assertEqual(inst.Meta.validator_class, self.validator_class)

    def test_init_with_validator_class_in_meta(self):
        class ConcreteSerializer(BaseValidatedSerializer):
            class Meta:
                validator_class = self.validator_class

        inst = generate_serializer(self.validator_class)()
        self.assertEqual(inst.Meta.validator_class, self.validator_class)

    def test_create_no_validation(self):
        class ConcreteValidator(Validator):
            pass

        data = {'somestring': 'a cool string'}
        self.assertEqual(serialize(ConcreteValidator, data)['somestring'], 'a cool string')

    def test_create_with_validation_clean(self):
        class ConcreteValidator(Validator):
            def clean(self, attrs: dict):
                attrs['somestring'] = attrs['somestring'].upper()

        data = {'somestring': 'a cool string'}
        self.assertEqual(serialize(ConcreteValidator, data)['somestring'], 'A COOL STRING')

    def test_create_with_validation_base(self):
        class ConcreteValidator(Validator):
            def base(self, attrs: dict):
                invalid_if(len(attrs['somestring']) > 5, 'somestring', 'too long')

            def will_create(self, attrs: dict):
                raise ArithmeticError('should stop at base')

            def base_db(self, attrs: dict):
                raise ArithmeticError('should stop at base')

        data = {'somestring': 'a cool string'}
        with self.assertRaises(serializers.ValidationError):
            serialize(ConcreteValidator, data)

    def test_create_with_validation_will_create(self):
        class ConcreteValidator(Validator):
            def will_create(self, attrs: dict):
                invalid_if(len(attrs['somestring']) > 5, 'somestring', 'too long')

        data = {'somestring': 'a cool string'}
        with self.assertRaises(serializers.ValidationError):
            serialize(ConcreteValidator, data)

    def test_create_with_validation_base_db(self):
        class ConcreteValidator(Validator):
            def base_db(self, attrs: dict):
                invalid('somestring', 'already in db')

        data = {'somestring': 'a cool string'}
        with self.assertRaises(serializers.ValidationError):
            serialize(ConcreteValidator, data)

    def test_create_with_validation_did_create(self):
        class ConcreteValidator(Validator):
            def clean(self, attrs: dict):
                if_in('somestring', attrs, lambda a: a.upper())

            def did_create(self, instance, validated_attrs: dict):
                assert 'somestring' in instance
                assert instance['somestring'] == 'A COOL STRING'
                assert validated_attrs['somestring'] == 'A COOL STRING'
                raise serializers.ValidationError('reached')

        data = {'somestring': 'a cool string'}
        with self.assertRaises(serializers.ValidationError) as ex:
            serialize(ConcreteValidator, data)
        self.assertEqual(ex.exception.detail[0], 'reached')

    def test_create_with_validation_to_representation(self):
        class ConcreteValidator(Validator):
            def to_representation(self, repr_attrs, validated_attrs=None):
                repr_attrs['something'] = 1
                return repr_attrs

        data = {'somestring': 'a cool string'}
        serializer = generate_serializer(ConcreteValidator)(data=data)
        serializer.is_valid(raise_exception=True)
        self.assertTrue('something' in serializer.data)
        self.assertEqual(serializer.data['something'], 1)


class ValidatedModelSerializerTests(TestCase):
    def setUp(self):
        class ModelAValidator(Validator):
            def __init__(self, **kwargs):
                super().__init__(**kwargs)
                assert hasattr(self, 'model')
                assert self.model == ModelA

            def clean(self, attrs: dict):
                attrs['field_1'] = 'a' + attrs['field_1']

            def base(self, attrs: dict):
                attrs['field_1'] = 'a' + attrs['field_1']

            def field_field_1(self, value):
                invalid_if('c' in value, 'field_1', 'Boom')

            def check_something(self, attrs):
                invalid_if('d' in attrs['field_1'], 'field_1', 'Boom')

            def will_create(self, attrs: dict):
                invalid_if(self.model.objects.count(), '', '')

        class ModelASerializer(ValidatedModelSerializer):
            class Meta:
                validator_class = ModelAValidator
                model = ModelA
                fields = '__all__'

        self.ModelAValidator = ModelAValidator
        self.ModelASerializer = ModelASerializer

    def test_create(self):
        with patch.object(self.ModelAValidator, 'will_create', return_value=None) as m1:
            with patch.object(self.ModelAValidator, 'did_create', return_value=None) as m2:
                serializer = self.ModelASerializer(data={'field_1': 'somestr', 'field_2': 15})
                serializer.is_valid(raise_exception=True)
                serializer.save()

        self.assertTrue(m1.called and m2.called)

        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().field_1, 'aasomestr')
        self.assertEqual(ModelA.objects.first().field_2, 15)

    def test_update(self):
        instance = ModelA.objects.create(field_1='otherstr', field_2=25)

        with patch.object(self.ModelAValidator, 'will_update', return_value=None) as m1:
            with patch.object(self.ModelAValidator, 'did_update', return_value=None) as m2:
                serializer = self.ModelASerializer(instance, data={'field_1': 'somestr', 'field_2': 15})
                serializer.is_valid(raise_exception=True)
                serializer.save()

        self.assertTrue(m1.called and m2.called)

        self.assertEqual(ModelA.objects.count(), 1)
        self.assertEqual(ModelA.objects.first().id, instance.id)
        self.assertEqual(ModelA.objects.first().field_1, 'aasomestr')
        self.assertEqual(ModelA.objects.first().field_2, 15)

    def test_validate_fields_positive(self):
        instance = ModelA.objects.create(field_1='otherstr', field_2=25)
        with patch.object(self.ModelASerializer, 'validate_extras', return_value=None) as m1:
            serializer = self.ModelASerializer(instance, data={'field_1': 'somestr', 'field_2': 15})
            serializer.is_valid(raise_exception=True)
            serializer.save()

        self.assertTrue(m1.called)

    def test_validate_fields_negative(self):
        instance = ModelA.objects.create(field_1='otherstr', field_2=25)
        serializer = self.ModelASerializer(instance, data={'field_1': 'somecstr', 'field_2': 15})
        with self.assertRaises(serializers.ValidationError) as ex:
            serializer.is_valid(raise_exception=True)
        self.assertEqual(ex.exception.detail['field_1'][0], 'Boom.')

    def test_validate_dependent_negative(self):
        instance = ModelA.objects.create(field_1='otherstr', field_2=25)
        serializer = self.ModelASerializer(instance, data={'field_1': 'somedstr', 'field_2': 15})
        with self.assertRaises(serializers.ValidationError) as ex:
            serializer.is_valid(raise_exception=True)
        self.assertEqual(ex.exception.detail['field_1'][0], 'Boom.')


from django.http import Http404

from django_alt.abstract.endpoints import MetaEndpoint
from django_alt.utils.shortcuts import queryset_has_many


class Endpoint(metaclass=MetaEndpoint):
    @classmethod
    def as_view(cls, **kwargs):
        """
        Main entry point for a request-response process.
        Allows to use handler.as_view() directly in urlpatterns.
        """
        return cls.view.as_view(**kwargs)

    """
    Default view handler implementations
    """

    @classmethod
    def on_get(cls, request, queryset, permission_test=None, **url) -> (dict, int):
        """
        Default GET handler implementation.
        Must return a tuple containing the response that is fed to the serializer and a status code.
        Safe to raise Validation, Permission and HTTP errors
        :param request: view request object
        :param queryset: queryset from the endpoint config
        :param permission_test: (optional) permission test to execute after full validation
        :param url: (optional) view url kwargs
        :return: {response_to_serialize, status_code}
        """
        if permission_test:
            cls.serializer._check_permissions(permission_test, request.data)
        return cls.serializer(queryset, many=queryset_has_many(queryset)).data, 200

    @classmethod
    def on_post(cls, request, permission_test=None, **url) -> (dict, int):
        """
        Default POST handler implementation
        Must return a tuple containing the response that is fed to the serializer and a status code.
        Safe to raise Validation, Permission and HTTP errors
        :param request: view request object
        :param permission_test: (optional) permission test to execute after full validation
        :param url: (optional) view url kwargs
        :return: {response_to_serialize, status_code}
        """
        is_many = isinstance(request.data, list)
        serializer = cls.serializer(data=request.data,
                                    permission_test=permission_test,
                                    many=is_many)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return serializer.data, 201

    @classmethod
    def on_patch(cls, request, queryset, permission_test=None, **url) -> (dict, int):
        """
        Default PATCH handler implementation
        Must return a tuple containing the response that is fed to the serializer and a status code.
        Safe to raise Validation, Permission and HTTP errors
        :param request: view request object
        :param queryset: queryset from the endpoint config
        :param [permission_test]: permission test to execute after full validation
        :param [url]: view url kwargs
        :return: {response_to_serialize, status_code}
        """
        if queryset is None:
            raise Http404
        if queryset_has_many(queryset):
            raise NotImplementedError()
        serializer = cls.serializer(queryset,
                                    data=request.data,
                                    many=queryset_has_many(queryset),
                                    partial=True,
                                    permission_test=permission_test)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return serializer.data, 200

    @classmethod
    def on_put(cls, request, queryset, permission_test=None, **url) -> (dict, int):
        """
        TBD
        :param request:
        :param queryset:
        :param permission_test:
        :param url:
        :return:
        """
        raise NotImplementedError()

    @classmethod
    def on_delete(cls, request, queryset, permission_test=None, **url) -> (dict, int):
        """
        Default DELETE handler implementation
        Must return a tuple containing the response that is fed to the serializer and a status code.
        Safe to raise Validation, Permission and HTTP errors
        :param request: view request object
        :param queryset: queryset from the endpoint config
        :param permission_test: (optional) permission test to execute after full validation
        :param url: (optional) view url kwargs
        :return: {response_to_serialize, status_code}
        """
        if queryset is None:
            raise Http404
        validator = cls.serializer.Meta.validator_class(model=cls.model)
        validator.will_delete(queryset)
        cls.serializer._check_permissions(permission_test, request.data)
        data = cls.serializer(queryset, many=queryset_has_many(queryset)).data
        queryset.delete()
        return data, 200

    """
    Default permission handler implementations
    """

    @classmethod
    def can_default(cls):
        """
        Defines a pair of optional permission checker functions (pre_validation_func, post_validation_func).
        First is called initially before any validation.
        Second is called after validation finishes.
        They return a `bool` value and their calling signatures are as defined below.
        Alternatively these values can be used instead of functions:
            True  -> permission always granted,
            False -> permission never granted,
            None  -> no permission needed (True <=> None, the distinction is only semantic)
        :return: (pre_validation_func=None, post_validation_func=None)
        """
        return (
            lambda request, **url: True,
            lambda request, queryset, attrs: True
        )

    @classmethod
    def can_get(cls):
        return cls.can_default()

    @classmethod
    def can_post(cls):
        return cls.can_default()

    @classmethod
    def can_patch(cls):
        return cls.can_default()

    @classmethod
    def can_put(cls):
        return cls.can_default()

    @classmethod
    def can_delete(cls):
        return cls.can_default()



from rest_framework import serializers

from .abstract.serializers import BaseValidatedSerializer


class ValidatedModelSerializer(BaseValidatedSerializer, serializers.ModelSerializer):
    def _instantiate_validator(self, **kwargs):
        assert hasattr(self.Meta, 'model'), (
            'Missing `model` field in serializer Meta class. '
            'Offending serializer: {0}').format(self.__class__.__name__)
        return self.Meta.validator_class(model=self.Meta.model, serializer=self, **kwargs)

    def create(self, validated_data: dict):
        instance = super().create(validated_data)
        self.validator.did_create(instance, validated_data)
        return instance

    def update(self, instance, validated_data: dict):
        instance = super().update(instance, validated_data)
        self.validator.did_update(instance, validated_data)
        return instance


class ValidatedModelListSerializer(serializers.ListSerializer):
    # def __init__(self, *args, **kwargs):
    #     pass

    def validate(self, attrs: dict):
        pass
        return attrs

    def create(self, validated_data: dict):
        raise NotImplementedError()

    def update(self, instance, validated_data: dict):
        raise NotImplementedError()



from django.db.models import QuerySet
from rest_framework import serializers

validation_error_class = serializers.ValidationError


def invalid_if(condition, key_or_list, error_or_list):
    """
    Shortcut for raising a validation error if a condition is met.
    :raises: serializers.ValidationError
    """
    if condition:
        invalid(key_or_list, error_or_list)


def invalid(key_or_list, error_or_list):
    """
    Shortcut for raising a validation error.
    :raises: serializers.ValidationError
    """
    raise validation_error_class(make_error(key_or_list, error_or_list))


def make_error(key_or_list, error_or_list) -> dict:
    """
    Creates a unified error object.
    :param key_or_list: field name (can be None if error is a non_field one or a list of str)
    :param error_or_list: one or more string representations of the error
    """
    err = error_or_list if isinstance(error_or_list, list) else [error_or_list]
    err = list(map(lambda e: (e if e[-1] == '.' or e[-1] == '!' else e + '.') if len(e) else e, err))
    if key_or_list:
        return {k: err for k in key_or_list} if isinstance(key_or_list, list) else {key_or_list: err}
    return {'non_field_errors': err}


def if_in(key, container, func_true=None, func_false=None):
    """
    Checks if given key on the container exists and executes.
    func_true or func_false with each corresponding value based on the outcome.
    :param key: the key to search
    :param container: the container to search in
    :param func_true: callable that is called on success: func_true(container[key])
    :param func_false: callable that is called on fail: func_false(container[key])
    :return: {bool} whether the key exists
    """
    if key in container and func_true is not None:
        container[key] = func_true(container[key])
        return True
    elif func_false is not None:
        container[key] = func_false()
        return False


def queryset_has_many(queryset) -> bool:
    """
    Checks whether a given query product is iterable.
    :param queryset: Django QuerySet object
    :return: {bool}
    """
    return isinstance(queryset, QuerySet) or hasattr(queryset, '__iter__')


def coal(obj, fallback):
    """
    Inspired by C# null-coalescing operator:
    C# --> not_null = obj ?? new Object()
    Py --> not_null = coal(obj, Object())
    Also returns fallback upon a KeyError
    """
    try:
        return obj if obj is not None else fallback
    except KeyError:
        return fallback


def coal_first(*args):
    """
    Finds the first argument that is not None and doesn't throw a KeyError
    :param args:
    :return:
    """
    for arg in args:
        if arg is not None:
            return arg
    return None


def if_all_in(keys, container, func_true=None):
    """
    checks if all given keys are in the container and
    executes func_true for each of the corresponding values
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether the keys exist
    """
    if all(key in container for key in keys):
        for key in keys:
            container[key] = func_true(container[key])
        return True
    return False


def if_any_in(keys, container, func_true=None):
    """
    checks if any given keys ar in the container and
    executes func_true for each of the corresponding values of the existing keys
    :param keys: iterable of keys to find
    :param container: the container to search in
    :param func_true: callable that is called for each value on success: func_true(container[key])
    :return: whether any of the keys exist
    """
    found_key = False
    for key in keys:
        if key in container:
            container[key] = func_true(container[key])
            found_key = True
    return found_key


def try_cast(typ, value):
    """
    Attempts to cast value to a given type.
    :param typ: type to cast to
    :param value: value to be casted
    :return: casted value if cast was successful, otherwise None
    """
    try:
        return typ(value)
    except ValueError:
        return None


from unittest import TestCase

from django_alt.utils.shortcuts import make_error, validation_error_class, invalid, invalid_if, coal


class UtilsShortcutsTests(TestCase):
    def test_make_error(self):
        self.assertEqual(make_error('k', 'v'), {'k': ['v.']})
        self.assertEqual(make_error('k', 'v.'), {'k': ['v.']})
        self.assertEqual(make_error(['k'], 'v'), {'k': ['v.']})
        self.assertEqual(make_error(['k'], ['v']), {'k': ['v.']})
        self.assertEqual(make_error(['k1', 'k2'], ['v']), {'k1': ['v.'], 'k2': ['v.']})
        self.assertEqual(make_error(['k1', 'k2'], ['v1', 'v2']), {'k1': ['v1.', 'v2.'], 'k2': ['v1.', 'v2.']})
        self.assertEqual(make_error(None, 'v1'), {'non_field_errors': ['v1.']})

    def test_invalid(self):
        with self.assertRaises(validation_error_class) as ex:
            invalid('k', 'v')
        self.assertEqual(ex.exception.detail, {'k': ['v.']})

    def test_invalid_if(self):
        invalid_if(False, 'k', 'v')
        with self.assertRaises(validation_error_class) as ex:
            invalid_if(True, 'k', 'v')
        self.assertEqual(ex.exception.detail, {'k': ['v.']})

    def test_coal(self):
        self.assertDictEqual(coal(None, {}), {})
        self.assertDictEqual(coal(None, {'a': 5}), {'a': 5})
        self.assertEqual(coal(1, {'a': 5}), 1)
        self.assertEqual(coal(False, {'a': 5}), False)
        self.assertEqual(coal(False, None), False)
        self.assertEqual(coal(None, False), False)
        self.assertEqual(coal(None, None), None)



from abc import abstractmethod
from collections import OrderedDict


class Validator:
    """
    Abstract class that defines the basic lifecycle hooks and definition
    principles for its subclasses
    """

    def __init__(self, *, model=None, serializer=None, **context):
        """
        :param [model]: model class of the serialized object (if serialized by a ModelSerializer)
        :param [context]: any data that gets passed as serializer kwargs
        """
        self.model = model
        self.serializer = serializer
        self.context = context

    @abstractmethod
    def clean(self, attrs: dict) -> dict:
        """
        Base clean method. Executed before base validation.
        Use this for
        - value cleaning (lowercasing, normalization, etc.)
        - setting or generating default dependent values (like slugs)
        :param attrs: dict containing model object attribute values
        :return: modified attrs
        """
        return attrs

    @abstractmethod
    def base(self, attrs: dict) -> dict:
        """
        Executed before create and update methods.
        Use this for
        - raising validation errors independent from create or update
        :param attrs: dict containing model object attribute values
        :return: modified attrs
        """
        return attrs

    @abstractmethod
    def base_db(self, attrs: dict) -> dict:
        """
        Executed as last step of validation. As DB access can be considered
        expensive, this will run last, provided any other steps check out.
        Use this for
        - validation that requires database access
        - validation logic that is more expensive
        :param attrs: dict containing model object attribute values
        :return: modified attrs
        """
        return attrs

    @abstractmethod
    def will_create(self, attrs: dict) -> dict:
        """
        Called when a new instance is created from attrs.
        Use this for
        - placing validation logic that should be solely executed on record creation
        :param attrs: attrs to create the instance
        :return: modified attrs
        """
        return attrs

    @abstractmethod
    def will_update(self, instance, attrs: dict) -> dict:
        """
        Called when an existing instance is populated with new attrs.
        Use this for
        - placing validation logic that should be solely executed on record update
        :param instance: existing instance
        :param attrs: attrs to update the instance with
        :return: modified attrs
        """
        return attrs

    @abstractmethod
    def will_delete(self, queryset):
        """
        Called before a queryset is deleted
        :param queryset: items to be deleted
        """
        pass

    @abstractmethod
    def did_create(self, instance, validated_attrs: dict) -> None:
        """
        Called after a model instance is created
        :param instance: the created model instance
        :param validated_attrs: validated attrs used to create the instance
        :return: None
        """
        pass

    @abstractmethod
    def did_update(self, instance, validated_attrs: dict) -> None:
        """
        Called after a model instance is created
        :param instance: the updated model instance
        :param validated_attrs: validated attrs used to update the instance
        :return: None
        """
        pass

    @abstractmethod
    def did_delete(self, attrs: dict) -> None:
        """
        Called after a model instance is deleted
        :param attrs: dict of data passed from the request object
        :return: None
        """
        pass

    @abstractmethod
    def to_representation(self, repr_attrs: OrderedDict, validated_attrs: dict = None) -> OrderedDict:
        """
        Called when a queried object is transformed to a dict of primitives by underlying DRF.
        Use this for
        - post-processing display values before converting them to JSON
        :param repr_attrs: an OrderedDict that is composed by DRF
        :param validated_attrs: an dict containing attributes that were passed through validation functions
        :return: modified attrs OrderedDict
        """
        return repr_attrs



from django_alt.abstract.validators import Validator
from django_alt.endpoints import Endpoint
from django_alt.serializers import ValidatedModelSerializer
from django_alt.utils.shortcuts import invalid_if
from django_alt_tests.conf.models import ModelA


class ModelAValidator(Validator):
    pass

class ModelAValidator2(Validator):
    def to_representation(self, repr_attrs, validated_attrs: dict = None):
        repr_attrs['extra'] = 'this is an extra'
        invalid_if(repr_attrs['id'] == 2, 'error', 'msg')


class ModelASerializer(ValidatedModelSerializer):
    class Meta:
        model = ModelA
        validator_class = ModelAValidator
        fields = '__all__'


class ModelASerializer2(ValidatedModelSerializer):
    class Meta:
        model = ModelA
        validator_class = ModelAValidator2
        fields = '__all__'


class ModelAEndpoint1(Endpoint):
    serializer = ModelASerializer
    config = {'get': dict(query=lambda model, **url: model.objects.all())}


class ModelAEndpoint2(Endpoint):
    serializer = ModelASerializer2
    config = {'get': dict(query=lambda model, **url: model.objects.all())}


class ModelAEndpoint3(Endpoint):
    serializer = ModelASerializer
    config = {'get': {
        'query': lambda model, **url: model.objects.all(),
        'filters': {
            'filter1': lambda queryset, param: queryset.filter(field_1=param),
            'filter2': lambda queryset, param: queryset.filter(field_2=param),
        }
    }}


class ModelAEndpoint4(Endpoint):
    serializer = ModelASerializer
    config = {'get': {
        'query': lambda model, **url: model.objects.all(),
        'filters': {
            'filter1': lambda queryset, param: queryset.filter(field_1=param),
            'filter2': lambda queryset, param: queryset.filter(field_2=param),
        }
    }}

    @classmethod
    def can_get(cls):
        return lambda request, **kwargs: False, None


class ModelAEndpoint5(Endpoint):
    serializer = ModelASerializer
    config = {'get': {
        'query': lambda model, **url: model.objects.all(),
        'filters': {
            'filter1': lambda queryset, param: queryset.filter(field_1=param),
            'filter2': lambda queryset, param: queryset.filter(field_2=param),
        }
    }}

    @classmethod
    def can_get(cls):
        return None, lambda request, qs, attrs: False


class ModelAEndpoint6(Endpoint):
    serializer = ModelASerializer
    config = {
        'post': None,
        'patch': {
            'query': lambda model_a, **_: model_a.objects.first()
        }
    }


class ModelAEndpoint7(Endpoint):
    serializer = ModelASerializer
    config = {
        'delete': {
            'query': lambda model_a, **_: model_a.objects.first()
        }
    }

class ModelAEndpoint8(Endpoint):
    serializer = ModelASerializer
    config = {
        'delete': {
            'query': lambda model_a, **_: model_a.objects.all()
        }
    }

class ModelAEndpoint9(Endpoint):
    serializer = ModelASerializer
    config = {
        'post': {
            'fields_from_url': ('field_1', 'field_2')
        }
    }

class ModelAEndpoint10(Endpoint):
    serializer = ModelASerializer
    config = {
        'post': {
            'fields_from_url': ('field_1', 'field_2', 'nonexistent_field')
        }
    }


from collections import OrderedDict

from django.core.exceptions import ObjectDoesNotExist

from rest_framework import serializers
from rest_framework.fields import empty

from django_alt.abstract.validators import Validator
from django_alt.utils.shortcuts import coal


class BaseValidatedSerializer(serializers.Serializer):
    """
    Defines a relation between rest_framework serializer and validator
    """
    FIELD_VALIDATOR_PREFIX = 'field_'
    ATTRS_VALIDATOR_PREFIX = 'check_'

    def __init__(self, instance=None, data=empty, *, validator_class=None, **kwargs):
        if not hasattr(self, 'Meta'):
            self.Meta = type('Meta', tuple(), dict())
        if validator_class:
            self.Meta.validator_class = validator_class

        assert hasattr(self.Meta, 'validator_class'), (
            'Either a `validator_class` set on the serializer Meta or '
            'provided when initializing the serializer is required.'
            'Offending serializer: {0}'
        ).format(self.__class__.__qualname__)

        self.permission_test = kwargs.pop('permission_test', None)
        self.did_check_permission = False

        self.Meta.validator_instance = self._instantiate_validator(**kwargs)
        super().__init__(instance, data, **kwargs)

    def _instantiate_validator(self, **kwargs):
        """
        Annotate the serializer with an instantiated model validator class
        :return: None
        """
        return self.Meta.validator_class(serializer=self, **kwargs)

    @staticmethod
    def _check_permissions(permission_test, attrs):
        try:
            if permission_test is not True and permission_test is not None and not permission_test(attrs):
                raise PermissionError()
        except ObjectDoesNotExist:
            raise PermissionError()
        return True

    @property
    def validator(self) -> Validator:
        """
        Fetches the validator instance bound to the serializer.
        The implementation of this behaviour varies for different
        serializer subclasses.
        :return: validator instance
        """
        return self.Meta.validator_instance

    @property
    def is_update(self) -> bool:
        """
        Helper method that checks if the serializer was initialized
        with an instance attribute to update
        :return: {bool}
        """
        return hasattr(self, 'instance') and self.instance is not None

    def validate_extras(self, attrs: dict):
        """
        If subclass defines functions named field_<field_name>
        where field_name corresponds to a field declared on the
        serializer, executes such functions with the field's value
        as the parameter.
        If subclass defines functions with names starting with check_,
        executes such functions with attrs dict as the parameter.
        All functions are called in alphabetical order
        :return: None
        """
        fields = sorted(self.fields.fields.keys())
        for field in fields:
            name = self.FIELD_VALIDATOR_PREFIX + field
            if hasattr(self.validator, name) and callable(getattr(self.validator, name)) and field in attrs:
                getattr(self.validator, name)(attrs[field])

        def is_attr_action(name):
            return name.startswith(self.ATTRS_VALIDATOR_PREFIX) and callable(getattr(self.validator, name))
        for name in [name for name in dir(self.validator) if is_attr_action(name)]:
            getattr(self.validator, name)(attrs)

    def check_permissions(self, attrs):
        """
        Performs post-permission checking, if a `permission_test` callable
        was passed to the serializer via its constructor
        :return: None
        :raises: PermissionError
        """
        self.permission_test is not None and self._check_permissions(self.permission_test, attrs)

    def validate(self, attrs: dict) -> dict:
        """
        Performs attribute validation before passing them to
        model managers.
        :param attrs: a dictionary containing input attributes to validate
        :return: transformed (if necessary) attributes from input
        """
        attrs = coal(self.validator.clean(attrs), attrs)
        attrs = coal(self.validator.base(attrs), attrs)

        self.validate_extras(attrs)

        if not self.is_update:
            attrs = coal(self.validator.will_create(attrs), attrs)
        else:
            attrs = coal(self.validator.will_update(self.instance, attrs), attrs)

        attrs = coal(self.validator.base_db(attrs), attrs)

        self.check_permissions(attrs)

        return attrs

    def to_representation(self, instance) -> OrderedDict:
        representation = super().to_representation(instance)
        result = self.validator.to_representation(representation,
                                                  self.validated_data if hasattr(self, '_validated_data') else None)
        return result if result else representation



import json
from functools import partial

from django.core.exceptions import ValidationError as DjangoValidationError, ObjectDoesNotExist
from django.http import Http404
from rest_framework import serializers
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from django_alt.utils.shortcuts import invalid, try_cast, coal_first

base_view_class = APIView
http_methods = ('get', 'post', 'patch', 'put', 'delete')

KW_CONFIG_FILTERS = 'filters'
KW_CONFIG_QUERYSET = 'query'
KW_CONFIG_URL_FIELDS = 'fields_from_url'
KW_CONFIG_URL_DONT_NORMALIZE = 'no_url_param_casting'


def _apply_filters(qs, filters, query_params):
    current_param = None
    try:
        for param, func in filters.items():
            current_param = param
            if param in query_params:
                qs = func(qs, query_params[param])
        return qs
    except DjangoValidationError as e:
        invalid(current_param, e.message)
    except ValueError as e:
        invalid(current_param, str(e))


def _normalize_url(**url):
    def cast(value):
        return coal_first(
            try_cast(int, value),
            try_cast(float, value),
            value
        )
    return {k: cast(v) for k, v in url.items()}


def _view_prototype(view_self, request, **url):
    method = request.method.lower()
    if method == 'head':
        method = 'get'
    endpoint = view_self.endpoint_class

    try:
        config = endpoint.config[method]

        if KW_CONFIG_URL_DONT_NORMALIZE not in config:
            url = _normalize_url(**url)

        if KW_CONFIG_URL_FIELDS in config:
            try:
                request.data.update({k: url[k] for k in config[KW_CONFIG_URL_FIELDS]})
            except KeyError as e:
                raise AssertionError(('Key supplied in `{0}` was not present in the url dict at endpoint `{1}`.\n'
                                      '`{0}` dump: {2}').format(KW_CONFIG_URL_FIELDS, view_self.endpoint_class.__name__,
                                                                json.dumps(config[KW_CONFIG_URL_FIELDS])))

        qs = None
        pre_can, post_can = getattr(endpoint, 'can_' + method)()

        # pre_can/post_can meaning
        if pre_can is not None and pre_can is not True:
            if pre_can is False or not pre_can(request, **url):
                raise PermissionError()

        if KW_CONFIG_QUERYSET in config:
            qs = config[KW_CONFIG_QUERYSET](endpoint.model, **url)
            if KW_CONFIG_FILTERS in config and len(request.query_params):
                qs = _apply_filters(qs, config[KW_CONFIG_FILTERS], request.query_params)

        if post_can is not None and post_can is not True:
            post_can = partial(post_can, request, qs)
        handler = getattr(endpoint, 'on_' + method)
        if method == 'post':
            return Response(*handler(request, post_can, **url))
        return Response(*handler(request, qs, post_can, **url))

    except serializers.ValidationError as e:
        return Response(e.detail, status=status.HTTP_400_BAD_REQUEST)

    except PermissionError:
        return Response(status=status.HTTP_401_UNAUTHORIZED if request.user.is_anonymous else status.HTTP_403_FORBIDDEN)

    except Http404:
        return Response(status=status.HTTP_404_NOT_FOUND)

    except (ObjectDoesNotExist, DjangoValidationError) as e:
        return Response(' '.join(e.args), status=status.HTTP_400_BAD_REQUEST)


class MetaEndpoint(type):
    def __new__(mcs, name, bases, clsdict):
        if len(bases):
            mcs.transform_fields(name, clsdict)
            if len(clsdict['config']):
                clsdict['view'] = mcs.make_view_class(name + 'View', clsdict['config'])
        cls = super().__new__(mcs, name, bases, clsdict)
        if len(bases) and len(clsdict['config']):
            cls.view.endpoint_class = cls
        return cls

    @staticmethod
    def transform_config_shorthands(config_dict: dict) -> dict:
        result = {}
        for k, v in config_dict.items():
            if v is None:
                v = {}
            if ',' in k:
                for end in k.split(','):
                    result.setdefault(end.strip(), {}).update(v)
            else:
                result.setdefault(k.strip(), {}).update(v)
        return result

    @staticmethod
    def transform_fields(name, clsdict):
        assert 'serializer' in clsdict, (
            'Field `serializer` is required for an endpoint definition. '
            'Offending Endpoint `{0}`'
        ).format(name)

        if 'model' not in clsdict:
            assert hasattr(clsdict['serializer'].Meta, 'model'), (
                'Explicit model field is missing and trying to get '
                'the model from the serializer failed at endpoint `{0}`.'
            ).format(name)
            clsdict['model'] = clsdict['serializer'].Meta.model

        config = clsdict.get('config', {})
        if len(config):
            assert isinstance(config, dict), (
                'Endpoint `config` field must be of type `dict`. Offending endpoint `{0}`'
            ).format(name)
            config = MetaEndpoint.transform_config_shorthands(config)

            for method_name, contents in config.items():
                assert method_name in http_methods, (
                    'Endpoint `{0}` config definition contains an unknown HTTP method `{2}`. '
                    'Allowed methods are `{1}`.'
                ).format(name, http_methods, method_name)

                if isinstance(contents, dict):
                    if KW_CONFIG_QUERYSET in contents:
                        assert callable(contents['query']), (
                            '`{0}` field in Endpoint `config` must be a callable accepting '
                            'parameters: `model` and `**url` in endpoint `{1}`'
                        ).format(KW_CONFIG_QUERYSET, name)
                        if 'filters' in contents:
                            assert isinstance(contents[KW_CONFIG_FILTERS], dict), (
                                '`{0}` field must be of type `dict`, containing pairs of `filter_name`, '
                                '`filter_func(queryset, value)` in endpoint `{1}`'
                            ).format(KW_CONFIG_FILTERS, name)

                    elif method_name == 'delete':
                        raise AssertionError(('`config` for `delete` must include '
                                              '`{0}` field for endpoint `{1}`').format(KW_CONFIG_QUERYSET, name))

                    elif KW_CONFIG_FILTERS in contents:
                        raise AssertionError('`{0}` cannot be used without `{1}` in endpoint `{2}`'
                                             .format(KW_CONFIG_FILTERS, KW_CONFIG_QUERYSET, name))

                    if KW_CONFIG_URL_FIELDS in contents:
                        assert hasattr(contents[KW_CONFIG_URL_FIELDS], '__iter__'), (
                            '`{0}` config field must be an iterable in endpoint `{1}`'
                        ).format(KW_CONFIG_URL_FIELDS, name)

                    if KW_CONFIG_URL_DONT_NORMALIZE in contents:
                        if KW_CONFIG_URL_DONT_NORMALIZE is not True:
                            del contents[KW_CONFIG_URL_DONT_NORMALIZE]

        clsdict['config'] = config

    @staticmethod
    def make_view_class(name, config: dict):
        body = {method: _view_prototype for method, _ in config.items()}
        return type(name, (base_view_class,), body)



from django.conf.urls import url
from . import endpoints as e

urlpatterns = [
    url(r'^1$', e.ModelAEndpoint1.as_view(), name='e1'),
    url(r'^2$', e.ModelAEndpoint2.as_view(), name='e2'),
    url(r'^3$', e.ModelAEndpoint3.as_view(), name='e3'),
    url(r'^4$', e.ModelAEndpoint4.as_view(), name='e4'),
    url(r'^5$', e.ModelAEndpoint5.as_view(), name='e5'),
    url(r'^6$', e.ModelAEndpoint6.as_view(), name='e6'),
    url(r'^7$', e.ModelAEndpoint7.as_view(), name='e7'),
    url(r'^8$', e.ModelAEndpoint8.as_view(), name='e8'),
    url(r'^9/(?P<field_1>\w+)/(?P<field_2>[0-9]+)/$', e.ModelAEndpoint9.as_view(), name='e9'),
    url(r'^10/(?P<field_1>\w+)/(?P<field_2>[0-9]+)/$', e.ModelAEndpoint10.as_view(), name='e10'),
]



from setuptools import setup, find_packages

setup(name='django-alt',
      version='0.60a0',
      description='Alternative approach to data validation and REST endpoints in Django and DRF',
      url='https://github.com/poskadesign/django-alt',
      author='Vilius Poška',
      author_email='vilius@poska.lt',
      license='MIT',
      install_requires=['django', 'djangorestframework'],
      packages=find_packages() + ['django_alt_tests', 'django_alt_tests.conf', 'django_alt_tests.conf.migrations'])
